"""Collection of utilities for performing commmon numerical computations in kinematic analysis."""
from biokinepy import (absor, cs, euler_angles, filling, segment_cs, joint_center, mean_smoother, np_utils, trajectory,
                       vec_ops, vel_acc)

__all__ = [
    'absor',
    'cs',
    'euler_angles',
    'filling',
    'segment_cs',
    'joint_center',
    'mean_smoother',
    'np_utils',
    'trajectory',
    'vec_ops',
    'vel_acc'
]



import numpy as np
from biokinepy.vec_ops import extended_dot


def gamage(marker_cluster):
    """Compute joint center mean from the trajectories of the tracking markers over time in the lab coordinate system
    (K, N, 3).

    Gamage, S. S. H. U. and J. Lasenby (2002). "New least squares solutions for estimating the average centre of
    rotation and the axis of rotation." J Biomech 35(1): 87-93.
    """
    num_markers = marker_cluster.shape[0]
    num_frames = marker_cluster.shape[1]
    a = np.zeros((3, 3))
    b = np.zeros((3, 1))
    for p in range(num_markers):
        present_idx = ~np.any(np.isnan(marker_cluster[p]), axis=1)
        marker_pos = marker_cluster[p][present_idx]
        if marker_pos.size == 0:
            continue
        a_1 = (marker_pos.T @ marker_pos) / num_frames
        v_p_hat = np.mean(marker_pos, axis=0)[..., np.newaxis]
        a_2 = v_p_hat @ v_p_hat.T
        a = a + 2 * (a_1 - a_2)
        v_p_2 = extended_dot(marker_pos, marker_pos)[..., np.newaxis]
        v_p_3 = marker_pos * v_p_2
        v_p_2_hat = np.mean(v_p_2)
        v_p_3_hat = np.mean(v_p_3, axis=0)[..., np.newaxis]
        b = b + v_p_3_hat - v_p_hat * v_p_2_hat
    return np.linalg.solve(a, b)



"""This module contains various useful numpy utilities."""

import numpy as np
from typing import Tuple


# written by alimanfoo and retrieved from:
# https://gist.github.com/alimanfoo/c5977e87111abe8127453b21204c1065
def find_runs(x: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Find runs of consecutive items in an array."""

    # ensure array
    x = np.asanyarray(x)
    if x.ndim != 1:
        raise ValueError('only 1D array supported')
    n = x.shape[0]

    # handle empty array
    if n == 0:
        return np.array([]), np.array([]), np.array([])

    else:
        # find run starts
        loc_run_start = np.empty(n, dtype=bool)
        loc_run_start[0] = True
        np.not_equal(x[:-1], x[1:], out=loc_run_start[1:])
        run_starts = np.nonzero(loc_run_start)[0]

        # find run values
        run_values = x[loc_run_start]

        # find run lengths
        run_lengths = np.diff(np.append(run_starts, n))

        return run_values, run_starts, run_lengths


def rms(vec: np.ndarray, axis: int = 0) -> np.ndarray:
    """Compute the RMS of vec."""
    return np.sqrt(np.mean(np.square(vec), axis=axis))


def nanrms(vec: np.ndarray, axis: int = 0) -> np.ndarray:
    """Compute the RMS of vec, ignoring NaN values."""
    return np.sqrt(np.nanmean(np.square(vec), axis=axis))


def mae(vec: np.ndarray, axis: int = 0) -> np.ndarray:
    """Compute the MAE (mean absolute error) of vec."""
    return np.mean(np.absolute(vec), axis=axis)


def nanmae(vec: np.ndarray, axis=0) -> np.ndarray:
    """Compute the MAE (mean absolute error) of vec, ignoring NaN values."""
    return np.nanmean(np.absolute(vec), axis=axis)



from typing import Tuple
import numpy as np


def extract_mat_components(mat: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray,
                                                     np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Return a mask indicating with frames have data, and the components of the rotation matrix individually."""
    not_nan = ~np.any(np.isnan(mat), (-2, -1))

    # extract matrix components
    r00 = mat[not_nan, 0, 0]
    r01 = mat[not_nan, 0, 1]
    r02 = mat[not_nan, 0, 2]

    r10 = mat[not_nan, 1, 0]
    r11 = mat[not_nan, 1, 1]
    r12 = mat[not_nan, 1, 2]

    r20 = mat[not_nan, 2, 0]
    r21 = mat[not_nan, 2, 1]
    r22 = mat[not_nan, 2, 2]

    return not_nan, r00, r01, r02, r10, r11, r12, r20, r21, r22


def zxy_intrinsic(mat: np.ndarray) -> np.ndarray:
    """Return zxy intrinsic Euler angle decomposition of mat (.., 4, 4))"""
    # extract components
    not_nan, r00, r01, r02, _, r11, _, r20, r21, r22 = extract_mat_components(mat)

    # pre-initialize results
    theta_z = np.full(not_nan.shape, np.nan)
    theta_x = np.full(not_nan.shape, np.nan)
    theta_y = np.full(not_nan.shape, np.nan)

    # compute Euler angles
    theta_z[not_nan] = np.where(r21 < 1, np.where(r21 > -1, np.arctan2(-r01, r11), -np.arctan2(r02, r00)),
                                np.arctan2(r02, r00))
    theta_x[not_nan] = np.where(r21 < 1, np.where(r21 > -1, np.arcsin(r21), -np.pi/2), np.pi/2)
    theta_y[not_nan] = np.where(r21 < 1, np.where(r21 > -1, np.arctan2(-r20, r22), 0), 0)

    return np.stack((theta_z, theta_x, theta_y), -1)


def yxz_intrinsic(mat: np.ndarray) -> np.ndarray:
    """Return yxz intrinsic Euler angle decomposition of mat (.., 4, 4))"""
    # extract components
    not_nan, r00, r01, r02, r10, r11, r12, _, _, r22 = extract_mat_components(mat)

    # pre-initialize results
    theta_y = np.full(not_nan.shape, np.nan)
    theta_x = np.full(not_nan.shape, np.nan)
    theta_z = np.full(not_nan.shape, np.nan)

    # compute Euler angles
    theta_y[not_nan] = np.where(r12 < 1, np.where(r12 > -1, np.arctan2(r02, r22), -np.arctan2(-r01, r00)),
                                np.arctan2(-r01, r00))
    theta_x[not_nan] = np.where(r12 < 1, np.where(r12 > -1, np.arcsin(-r12), np.pi/2), -np.pi/2)
    theta_z[not_nan] = np.where(r12 < 1, np.where(r12 > -1, np.arctan2(r10, r11), 0), 0)

    return np.stack((theta_y, theta_x, theta_z), -1)


def yxy_intrinsic(mat: np.ndarray) -> np.ndarray:
    """Return yxy intrinsic Euler angle decomposition of mat (.., 4, 4))"""
    # extract components
    not_nan, r00, r01, r02, r10, r11, r12, _, r21, _ = extract_mat_components(mat)

    # pre-initialize results
    theta_y0 = np.full(not_nan.shape, np.nan)
    theta_x = np.full(not_nan.shape, np.nan)
    theta_y1 = np.full(not_nan.shape, np.nan)

    # compute Euler angles
    theta_y0[not_nan] = np.where(r11 < 1, np.where(r11 > -1, np.arctan2(-r01, -r21), 0), 0)
    theta_x[not_nan] = np.where(r11 < 1, np.where(r11 > -1, -np.arccos(r11), -np.pi), 0)
    theta_y1[not_nan] = np.where(r11 < 1, np.where(r11 > -1, np.arctan2(-r10, r12), np.arctan2(r02, r00)),
                                 np.arctan2(r02, r00))

    return np.stack((theta_y0, theta_x, theta_y1), -1)


def xzy_intrinsic(mat: np.ndarray) -> np.ndarray:
    """Return xzy intrinsic Euler angle decomposition of mat (.., 4, 4))"""
    # extract components
    not_nan, r00, r01, r02, _, r11, _, r20, r21, r22 = extract_mat_components(mat)

    # pre-initialize results
    theta_x = np.full(not_nan.shape, np.nan)
    theta_z = np.full(not_nan.shape, np.nan)
    theta_y = np.full(not_nan.shape, np.nan)

    # compute Euler angles
    theta_x[not_nan] = np.where(r01 < 1, np.where(r01 > -1, np.arctan2(r21, r11), -np.arctan2(-r20, r22)),
                                np.arctan2(-r20, r22))
    theta_z[not_nan] = np.where(r01 < 1, np.where(r01 > -1, np.arcsin(-r01), np.pi/2), -np.pi/2)
    theta_y[not_nan] = np.where(r01 < 1, np.where(r01 > -1, np.arctan2(r02, r00), 0), 0)

    return np.stack((theta_x, theta_z, theta_y), -1)


def gh_isb(mat: np.ndarray) -> np.ndarray:
    """Return GH ISB (yxy intrinsic) decomposition."""
    return yxy_intrinsic(mat)


def gh_phadke(mat: np.ndarray) -> np.ndarray:
    """Return GH Phadke (xzy intrinsic) decomposition."""
    return xzy_intrinsic(mat)


def ht_isb(mat: np.ndarray) -> np.ndarray:
    """Return HT ISB (yxy intrinsic) decomposition."""
    return yxy_intrinsic(mat)


def ht_phadke(mat: np.ndarray) -> np.ndarray:
    """Return HT Phadke (xzy intrinsic) decomposition."""
    return xzy_intrinsic(mat)


def st_isb(mat: np.ndarray) -> np.ndarray:
    """Return ST ISB (yxz intrisc) decomposition."""
    return yxz_intrinsic(mat)


def thorax_isb(mat: np.ndarray) -> np.ndarray:
    """Return thorax relative to global ISB (zxy intrinsic) decomposition."""
    return zxy_intrinsic(mat)


gh_isb.legend = ['Plane of Elevation', 'Elevation(-)', 'Axial Rotation I(+)/E(-)']
ht_isb.legend = ['Plane of Elevation', 'Elevation(-)', 'Axial Rotation I(+)/E(-)']
gh_phadke.legend = ['Elevation(-)', 'Angle of Flexion(+)/Horizontal Abd(-)', 'Axial Rotation I(+)/E(-)']
ht_phadke.legend = ['Elevation(-)', 'Angle of Flexion(+)/Horizontal Abd(-)', 'Axial Rotation I(+)/E(-)']
st_isb.legend = ['Retraction(+)/Protraction(-)', 'Lateral(-)/Medial(+) Rotation', 'Anterior(-)/Posterior(+) Tilt']
thorax_isb.legend = ['Flexion(+)/Extension(-)', 'Lateral Flexion L(+)/R(-)', 'Axial Rotation L(+)/R(-)']

gh_isb.legend_short = ['PoE', 'Elevation', 'Axial']
ht_isb.legend_short = ['PoE', 'Elevation', 'Axial']
gh_phadke.legend_short = ['Elevation', 'Flex/Abd', 'Axial']
ht_phadke.legend_short = ['Elevation', 'Flex/Abd', 'Axial']
st_isb.legend_short = ['Re/Protraction', 'Rotation', 'Tilt']
thorax_isb.legend_short = ['Flex/Ext', 'L/R Flex', 'Axial']



import numpy as np


def scapula_cs_isb(ac_gc: np.ndarray, ia: np.ndarray, ts: np.ndarray) -> np.ndarray:
    """Return scapula frame (ISB) given landmarks."""
    z_axis = ac_gc - ts
    x_axis = np.cross(z_axis, ia - ts)
    y_axis = np.cross(z_axis, x_axis)

    x_axis = x_axis/np.linalg.norm(x_axis)
    y_axis = y_axis/np.linalg.norm(y_axis)
    z_axis = z_axis/np.linalg.norm(z_axis)

    scap = np.eye(4)
    scap[0:3, 0] = x_axis
    scap[0:3, 1] = y_axis
    scap[0:3, 2] = z_axis
    scap[0:3, 3] = ac_gc

    return scap


import numpy as np
import quaternion
from lazy import lazy
from scipy.integrate import cumtrapz
from biokinepy.cs import ht_r, ht_inv, change_cs
from biokinepy.vec_ops import extended_dot
from biokinepy.vel_acc import ang_vel
from biokinepy.mean_smoother import smooth_pos_traj, smooth_quat_traj
import biokinepy.euler_angles as euler_angles


class EulerTrajectory:
    """An Euler angle trajectory."""
    def __init__(self, pose_trajectory: 'PoseTrajectory'):
        self.pose_trajectory = pose_trajectory

    def __getattr__(self, decomp_method):
        ret = getattr(euler_angles, decomp_method)(self.pose_trajectory.rot_matrix)
        self.__dict__[decomp_method] = ret
        return ret


class PoseTrajectory:
    """A rigid-body pose trajectory.

    Attributes
    ----------
    pos: np.ndarray (N, 3)
        Position trajectory.
    dt: float
        Frame period - expressed in seconds.
    frame_nums: np.ndarray (N, 1)
         Frame numbers associated with trajectory.
    """

    @classmethod
    def from_quat(cls, pos, quat, dt=1, frame_nums=None):
        """Create trajectory from scalar-first quaternion (N, 4) and position (N, 3) trajectory."""
        return cls(pos=pos, quat=quat, dt=dt, frame_nums=frame_nums)

    @classmethod
    def from_matrix(cls, pos, rot_mat, dt=1, frame_nums=None):
        """Create trajectory from rotation matrix (N, 3, 3) and position (N, 3) trajectory."""
        return cls(pos=pos, rot_mat=rot_mat, dt=dt, frame_nums=frame_nums)

    @classmethod
    def from_ht(cls, ht, dt=1, frame_nums=None):
        """Create trajectory from homogeneous transformation matrix (N, 4, 4) trajectory."""
        return cls(ht=ht, dt=dt, frame_nums=frame_nums)

    def __init__(self, ht=None, pos=None, quat=None, rot_mat=None, dt=1, frame_nums=None):
        self.frame_nums = frame_nums
        self.dt = dt
        self.long_axis = None

        # establish position
        assert((ht is not None) ^ (pos is not None))
        if ht is not None:
            self.__dict__['ht'] = ht
            self.__dict__['rot_matrix'] = ht[..., :3, :3]
            self.pos = ht[..., :3, 3]
        else:
            self.pos = pos

        # establish orientation
        assert((quat is not None) ^ (rot_mat is not None) ^ (ht is not None))
        if quat is not None:
            assert(pos.shape[0] == quat.shape[0])
            self.__dict__['quat_float'] = quat / np.sqrt(extended_dot(quat, quat))[..., np.newaxis]
            self.__dict__['quat'] = quaternion.from_float_array(self.__dict__['quat_float'])
        elif rot_mat is not None:
            assert (pos.shape[0] == rot_mat.shape[0])
            self.__dict__['rot_matrix'] = rot_mat
        else:
            # else ht has been passed which was handled above
            pass

    @lazy
    def quat(self) -> np.ndarray:
        """Return trajectory orientation as a scalar-first quaternion (numpy.quaternion) trajectory."""
        # if self.quat has been set in constructor this will never get called
        return quaternion.from_rotation_matrix(self.rot_matrix, nonorthogonal=False)

    @property
    def quat_float(self) -> np.ndarray:
        """Return trajectory orientation as a scalar-first quaternion (numpy.float) trajectory."""
        # if self.quat_float has been set in constructor this will never get called
        return quaternion.as_float_array(self.quat)

    @lazy
    def rot_matrix(self) -> np.ndarray:
        """Return trajectory orientation as a rotation matrix (N, 3, 3) trajectory."""
        # if self.rot_matrix has been set in constructor this will never get called
        return quaternion.as_rotation_matrix(self.quat)

    @lazy
    def ht(self) -> np.ndarray:
        """Return trajectory pose as a homogeneous transformation matrix (N, 4, 4) trajectory."""
        # if self.ht has been set in constructor this will never get called
        return ht_r(self.rot_matrix, self.pos)

    def in_frame(self, ht) -> 'PoseTrajectory':
        """Return a new pose trajectory which is the current trajectory expressed in the frame defined in ht."""
        return PoseTrajectory.from_ht(change_cs(ht_inv(ht), self.ht), self.dt, self.frame_nums)

    def in_trajectory(self, traj) -> 'PoseTrajectory':
        """Return a new pose trajectory which is the current trajectory expressed in the trajectory defined in ht."""
        return PoseTrajectory.from_ht(change_cs(ht_inv(traj.ht), self.ht), self.dt, self.frame_nums)

    @lazy
    def euler(self) -> EulerTrajectory:
        """Return an EulerTrajectory object which enables various types of Euler decompositions for the current
        trajectory."""
        return EulerTrajectory(self)

    @lazy
    def vel(self) -> np.ndarray:
        """Return trajectory linear velocity."""
        return np.gradient(self.pos, self.dt, axis=0)

    @lazy
    def ang_vel(self) -> np.ndarray:
        """Return trajectory angular velocity."""
        return ang_vel(self.rot_matrix, self.dt)

    @lazy
    def true_axial_rot(self) -> np.ndarray:
        """Return the true axial rotation of the trajectory."""
        if self.long_axis is None:
            raise ValueError('You must set the long_axis in order to compute true axial rotation.')
        ang_vel_proj = extended_dot(self.ang_vel, np.squeeze(self.rot_matrix @
                                                             self.long_axis[np.newaxis, ..., np.newaxis]))
        return cumtrapz(ang_vel_proj, dx=self.dt, initial=0)


def smooth_trajectory(traj: PoseTrajectory, num_frames_avg: int) -> PoseTrajectory:
    return PoseTrajectory.from_quat(smooth_pos_traj(traj.pos, num_frames_avg),
                                    smooth_quat_traj(traj.quat_float, num_frames_avg), traj.dt, traj.frame_nums)



from typing import Tuple
import numpy as np
import quaternion


def absor_matrix(markers_a: np.ndarray, markers_b: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Return optimal rigid transformation between markers_a (3, N) and markers_b (3, N).

    Söderkvist, I. and P.-Å. Wedin (1993). "Determining the movements of the skeleton using well-configured markers."
    J Biomech 26(12): 1473-1477.
    """
    a_mean = np.mean(markers_a, 1)
    b_mean = np.mean(markers_b, 1)
    pos_a_centered = markers_a - a_mean[:, np.newaxis]
    pos_b_centered = markers_b - b_mean[:, np.newaxis]
    u, s, vt = np.linalg.svd(pos_b_centered @ pos_a_centered.T)
    r = u @ (np.diag([1, 1, np.linalg.det(u @ vt)])) @ vt
    t = b_mean - r.dot(a_mean)
    return r, t


def absor_quat(markers_a: np.ndarray, markers_b: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Return optimal rigid transformation between markers_a (3, N) and markers_b (3, N).

    Horn, B. K. P. (1987). "Closed-form solution of absolute orientation using unit quaternions."
    Journal of the Optical Society of America A 4(4): 629-642.
    """
    a_mean = np.mean(markers_a, 1)
    b_mean = np.mean(markers_b, 1)
    pos_a_centered = markers_a - a_mean[:, np.newaxis]
    pos_b_centered = markers_b - b_mean[:, np.newaxis]
    m = pos_a_centered @ pos_b_centered.T

    # compute n matrix detailed in paper
    delta = np.array([m[1, 2] - m[2, 1], m[2, 0] - m[0, 2], m[0, 1] - m[1, 0]])
    n = np.zeros((4, 4))
    n[0, :] = np.array([np.trace(m), delta[0], delta[1], delta[2]])
    n[1:4, 0] = delta
    n[1:4, 1:4] = m + m.T - np.trace(m) * np.eye(3)

    # find optimal R and t
    (eigen_values, eigen_vectors) = np.linalg.eig(n)
    q = eigen_vectors[:, eigen_values.argmax()]
    r = quaternion.as_rotation_matrix(quaternion.from_float_array(q))
    t = b_mean - r.dot(a_mean)
    return r, t


def compute_trajectory(static_markers: np.ndarray, tracking_markers: np.ndarray) -> np.ndarray:
    """Return segment kinematic trajectory (N, 4, 4) given tracking markers in the static trial expressed in segment's
    coordinate system (K, 3) and the trajectories of the tracking markers over time in the lab
    coordinate system (K, N, 3)."""

    # markers_present (K, N)
    markers_present = ~np.any(np.isnan(tracking_markers), 2)
    num_frames = tracking_markers[0].shape[0]
    trajectory = np.full((num_frames, 4, 4), np.nan)

    for i in range(num_frames):
        # need at least 3 markers to determine the rigid transformation
        if np.count_nonzero(markers_present[:, i]) >= 3:
            # markers for this frame, removing markers with no data
            static_markers_frame = static_markers[markers_present[:, i], :]
            tracking_markers_frame = tracking_markers[markers_present[:, i], i, :]
            r, t = absor_matrix(static_markers_frame.T, tracking_markers_frame.T)
            trajectory[i, 0:3, 0:3] = r
            trajectory[i, 0:3, 3] = t
            trajectory[i, 3, :] = [0, 0, 0, 1]

    return trajectory


def compute_trajectory_continuous(static_markers: np.ndarray, tracking_markers: np.ndarray) -> np.ndarray:
    """Return segment kinematic trajectory (N, 4, 4) given tracking markers in the static trial expressed in segment's
    coordinate system (K, 3) and the trajectories of the tracking markers over time in the lab
    coordinate system (K, N, 3)."""

    # pre-initialize return value
    num_frames = tracking_markers[0].shape[0]
    trajectory = np.full((num_frames, 4, 4), np.nan)

    # remove markers that have no data at all
    no_data_markers = np.all(np.isnan(tracking_markers), (1, 2))
    data_markers = ~no_data_markers
    if np.count_nonzero(data_markers) < 3:
        return trajectory
    static_markers_wdata = static_markers[data_markers]
    tracking_markers_wdata = tracking_markers[data_markers]

    # find endpoints of cluster - do not compute kinematics prior to and post these endpoints
    cluster_present_frms = np.nonzero(~np.any(np.isnan(tracking_markers_wdata), (0, 2)))[0]

    for i in range(cluster_present_frms[0], cluster_present_frms[-1] + 1):
        r, t = absor_matrix(static_markers_wdata.T, tracking_markers_wdata[:, i, :].T)
        trajectory[i, 0:3, 0:3] = r
        trajectory[i, 0:3, 3] = t
        trajectory[i, 3, :] = [0, 0, 0, 1]

    return trajectory



"""This module contains utilities for performing common vector operations."""
import numpy as np


def extended_dot(vecs1: np.ndarray, vecs2: np.ndarray) -> np.ndarray:
    """Peform the dot product of the N vectors in vecs1 (N,x) with the corresponding vectors in vecs2 (N, x)."""

    # because the array dimensions must be changed to use matrix multiplication to perform this operation, einsum is
    # (as far as I know) the fastest way to perform this operation
    return np.einsum('ij,ij->i', vecs1, vecs2)



import numpy as np
from typing import Tuple, List
from biokinepy.np_utils import find_runs
from biokinepy.absor import absor_matrix


def fill_gaps_rb(marker_pos: np.ndarray, cluster_marker_pos: np.ndarray) -> Tuple[np.ndarray, List[Tuple[int, int]]]:
    """Rigid body fill the gaps present in marker_pos using cluster_marker_pos."""
    # find endpoints of cluster - we cannot fill beyond these
    cluster_present = np.nonzero(~np.any(np.isnan(cluster_marker_pos), (0, 2)))[0]
    first_frame_idx = cluster_present[0]
    last_frame_idx = cluster_present[-1]
    marker_present = ~np.any(np.isnan(marker_pos[first_frame_idx:last_frame_idx+1]), 1)
    run_values, run_starts, run_lengths = find_runs(marker_present)
    gaps_start = run_starts[~run_values] + first_frame_idx
    gaps_end = gaps_start + run_lengths[~run_values]

    marker_pos_filled = marker_pos.copy()
    gaps = list(zip(gaps_start, gaps_end))
    for gap_start, gap_end in gaps:
        assert (~np.any(np.isnan(cluster_marker_pos[:, gap_start:gap_end, :])))
        # if gap is at the beginning of the trial
        if gap_start == first_frame_idx:
            marker_pos_right = marker_pos[gap_end]
            for frame_num in range(gap_start, gap_end):
                r, t = absor_matrix(cluster_marker_pos[:, gap_end, :].T, cluster_marker_pos[:, frame_num, :].T)
                marker_pos_filled[frame_num] = r.dot(marker_pos_right) + t
        # if the gap is at the end of the trial
        elif gap_end == last_frame_idx + 1:
            marker_pos_left = marker_pos[gap_start-1]
            for frame_num in range(gap_start, gap_end):
                r, t = absor_matrix(cluster_marker_pos[:, gap_start-1, :].T, cluster_marker_pos[:, frame_num, :].T)
                marker_pos_filled[frame_num] = r.dot(marker_pos_left) + t
        # if the gap is in the middle
        else:
            marker_pos_left = marker_pos[gap_start-1]
            marker_pos_right = marker_pos[gap_end]
            for frame_num in range(gap_start, gap_end):
                total_frames = gap_end - (gap_start - 1)
                r_l, t_l = absor_matrix(cluster_marker_pos[:, gap_start-1, :].T, cluster_marker_pos[:, frame_num, :].T)
                r_r, t_r = absor_matrix(cluster_marker_pos[:, gap_end, :].T, cluster_marker_pos[:, frame_num, :].T)
                marker_pred_left = r_l.dot(marker_pos_left) + t_l
                marker_pred_right = r_r.dot(marker_pos_right) + t_r
                marker_pos_filled[frame_num] = (marker_pred_left * (gap_end-frame_num)/total_frames +
                                                marker_pred_right * (frame_num-(gap_start - 1))/total_frames)

    return marker_pos_filled, gaps



import numpy as np
import quaternion


def ang_vel(mat_traj: np.ndarray, dt) -> np.ndarray:
    """Return the angular velocity of the rotation matrix trajectory (N, 3, 3)."""
    mats_vel = np.gradient(mat_traj, dt, axis=0)
    mats_t = np.swapaxes(mat_traj, -2, -1)
    ang_vel_tensor = mats_vel @ mats_t
    ang_vel_vector = np.stack((ang_vel_tensor[:, 2, 1], ang_vel_tensor[:, 0, 2], ang_vel_tensor[:, 1, 0]), -1)
    return ang_vel_vector


def ang_vel_quat(quat_traj: np.ndarray, dt) -> np.ndarray:
    """Return the angular velocity of a quaternion trajectory (N, 4)."""
    mats = quaternion.as_rotation_matrix(quaternion.as_quat_array(quat_traj))
    return ang_vel(mats, dt)



"""This module provides utilities for performing common coordinate system operations."""
import numpy as np


def ht_r(r: np.ndarray, t: np.ndarray) -> np.ndarray:
    """Combine r (rotation) and t (translation) into a (..., 4, 4) homogeneous matrix (numpy ndarray)."""
    new_shape = list(r.shape)
    new_shape[-1] = new_shape[-1] + 1
    new_shape[-2] = new_shape[-2] + 1
    mat = np.zeros(new_shape)
    mat[..., 3, 3] = 1
    mat[..., :3, :3] = r
    mat[..., :3, 3] = t
    return mat


def ht_inv(mat: np.ndarray) -> np.ndarray:
    """Return the inverse of the homogeneous transformation in mat."""
    mat_inv = np.zeros_like(mat)
    mat_inv[..., 3, 3] = 1
    mat_inv[..., :3, :3] = np.swapaxes(mat[..., :3, :3], -1, -2)
    mat_inv[..., :3, 3] = np.squeeze(-mat_inv[..., :3, :3] @ mat[..., :3, 3, np.newaxis])
    return mat_inv


def vec_transform(t: np.ndarray, vecs: np.ndarray) -> np.ndarray:
    """Transform the N vectors in vecs (N, 4) using the homogeneous matrix (4, 4) t."""
    # (t @ vecs.T).T = vecs @ t.T
    return vecs @ t.T


def make_vec_hom(vec: np.ndarray) -> np.ndarray:
    """Transform the N vectors in vec (N, 3) into homogeneous counterparts (N, 4)."""
    return np.concatenate((vec, np.ones((vec.shape[0], 1))), axis=1)


def change_cs(new_frame: np.ndarray, frame_traj: np.ndarray) -> np.ndarray:
    """Change all frames in frame_traj (N, 4, 4) into the reference CS specified by new_frame (either (4, 4) or
    (N, 4, 4)."""
    return new_frame @ frame_traj


def pos_ht(mats: np.ndarray) -> np.ndarray:
    """Extract position from homogeneous transformations in mats (N, 4, 4)."""
    return mats[..., :3, 3]


def r_ht(mats: np.ndarray) -> np.ndarray:
    """Extract orientation from homogeneous transformations in mats (N, 4, 4)."""
    return mats[..., :3, :3]



import setuptools

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name='biokinepy',
    version='0.1.1',
    author='Klevis Aliaj',
    author_email='klevis.a@gmail.com',
    description='Utilities for processing, computing, and analyzing joint kinematics.',
    keywords=['biomechanics', 'kinematics', 'kinematic analysis', 'motion capture', 'skin marker', 'quaternion',
              'homogeneous transformation', 'angular velocity'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/klevis-a/biokinepy',
    license='MIT',
    packages=setuptools.find_packages(),
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],
    python_requires='>=3.7',
    install_requires=['numpy', 'numpy-quaternion', 'scipy', 'lazy']
)



import numpy as np
import quaternion


def quat_mean(quats: np.ndarray) -> np.ndarray:
    """Return average of quaternions in quats (N, 4)."""
    k = np.dot(quats.T, quats)
    l, v = np.linalg.eigh(k)
    return v[:, -1]


def smooth_quat_traj(quat_traj_floats: np.ndarray, num_frames_avg: int) -> np.ndarray:
    """Return moving average filtered trajectory of quaternions in quat_traj_floats (N, 4) - scalar-first convention."""
    assert(num_frames_avg % 2 == 1)
    num_frames_ext = int((num_frames_avg - 1) / 2)
    quat_traj = quaternion.as_quat_array(quat_traj_floats)

    # extend trajectory
    quat_left_ext = quat_traj[0] * quat_traj[num_frames_ext:0:-1].conjugate() * quat_traj[0]
    quat_right_ext = quat_traj[-1] * quat_traj[-2:-2 - num_frames_ext:-1].conjugate() * quat_traj[-1]
    quat_traj_ext = np.concatenate((quat_left_ext, quat_traj, quat_right_ext))

    # compute average
    quat_traj_smooth_floats = np.zeros_like(quat_traj_floats)
    for i in range(quat_traj_smooth_floats.shape[0]):
        q_mean = quat_mean(quaternion.as_float_array(quat_traj_ext[i:i + num_frames_avg]))
        # since q and -q represent the same orientation make sure that we point in the same general direction
        quat_traj_smooth_floats[i] = -q_mean if q_mean.dot(quat_traj_floats[i]) < 0 else q_mean

    return quat_traj_smooth_floats


def smooth_pos_traj(pos_traj: np.ndarray, num_frames_avg: int) -> np.ndarray:
    """Return moving average filtered position trajectory (N, 3)."""
    assert(num_frames_avg % 2 == 1)
    num_frames_ext = int((num_frames_avg - 1) / 2)

    # extend trajectory
    pos_left_ext = -(pos_traj[num_frames_ext:0:-1] - pos_traj[0]) + pos_traj[0]
    pos_right_ext = -(pos_traj[-2:-2 - num_frames_ext:-1] - pos_traj[-1]) + pos_traj[-1]
    pos_traj_ext = np.concatenate((pos_left_ext, pos_traj, pos_right_ext))

    # average stencil
    stenc = np.ones(num_frames_avg) / num_frames_avg

    # smooth
    pos_traj_smooth = np.empty_like(pos_traj)
    for i in range(3):
        pos_traj_smooth[:, i] = np.convolve(pos_traj_ext[:, i], stenc, 'valid')

    return pos_traj_smooth
